#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Block Storage Volume/Snapshot Management implementations"""

import argparse
from collections.abc import Iterable
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


SORT_MANAGEABLE_KEY_VALUES = ('size', 'reference')


class BlockStorageManageVolumes(command.Lister):
    """List manageable volumes.

    Supported by --os-volume-api-version 3.8 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        host_group = parser.add_mutually_exclusive_group()
        host_group.add_argument(
            "host",
            metavar="<host>",
            nargs='?',
            help=_(
                'Cinder host on which to list manageable volumes. '
                'Takes the form: host@backend-name#pool'
            ),
        )
        host_group.add_argument(
            "--cluster",
            metavar="<cluster>",
            help=_(
                'Cinder cluster on which to list manageable volumes. '
                'Takes the form: cluster@backend-name#pool. '
                '(supported by --os-volume-api-version 3.17 or later)'
            ),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        # TODO(stephenfin): Remove this in a future major version bump
        parser.add_argument(
            '--detailed',
            metavar='<detailed>',
            default=None,
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            '--marker',
            metavar='<marker>',
            default=None,
            help=_(
                'Begin returning volumes that appear later in the volume '
                'list than that represented by this reference. This '
                'reference should be json like. Default=None.'
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            default=None,
            help=_('Maximum number of volumes to return. Default=None.'),
        )
        parser.add_argument(
            '--offset',
            metavar='<offset>',
            default=None,
            help=_('Number of volumes to skip after marker. Default=None.'),
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            default=None,
            help=(
                _(
                    'Comma-separated list of sort keys and directions in the '
                    'form of <key>[:<asc|desc>]. '
                    'Valid keys: %s. '
                    'Default=None.'
                )
                % ', '.join(SORT_MANAGEABLE_KEY_VALUES)
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[list[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if parsed_args.host is None and parsed_args.cluster is None:
            msg = _(
                "Either <host> or '--cluster <cluster>' needs to be provided "
                "to run the 'block storage volume manageable list' command"
            )
            raise exceptions.CommandError(msg)

        if not sdk_utils.supports_microversion(volume_client, '3.8'):
            msg = _(
                "--os-volume-api-version 3.8 or greater is required to "
                "support the 'block storage volume manageable list' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.cluster:
            if not sdk_utils.supports_microversion(volume_client, '3.17'):
                msg = _(
                    "--os-volume-api-version 3.17 or greater is required to "
                    "support the '--cluster' option"
                )
                raise exceptions.CommandError(msg)

        detailed = parsed_args.long
        if parsed_args.detailed is not None:
            detailed = parsed_args.detailed.lower().strip() in {
                '1',
                't',
                'true',
                'on',
                'y',
                'yes',
            }
            if detailed:
                # if the user requested e.g. '--detailed true' then they should
                # not request '--long'
                msg = _(
                    "The --detailed option has been deprecated. "
                    "Use --long instead."
                )
                self.log.warning(msg)
            else:
                # if the user requested e.g. '--detailed false' then they
                # should simply stop requesting this since the default has
                # changed
                msg = _("The --detailed option has been deprecated. Unset it.")
                self.log.warning(msg)

        columns = [
            'reference',
            'size',
            'safe_to_manage',
        ]
        if detailed:
            columns.extend(
                [
                    'reason_not_safe',
                    'cinder_id',
                    'extra_info',
                ]
            )

        query: dict[str, Any] = {}
        if parsed_args.host:
            query['host'] = parsed_args.host
        if parsed_args.cluster:
            query['cluster'] = parsed_args.cluster
        if parsed_args.marker:
            query['marker'] = parsed_args.marker
        if parsed_args.limit:
            query['limit'] = parsed_args.limit
        if parsed_args.offset:
            query['offset'] = parsed_args.offset
        if parsed_args.sort:
            query['sort'] = parsed_args.sort

        data = volume_client.manageable_volumes(details=detailed, **query)

        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class BlockStorageManageSnapshots(command.Lister):
    """List manageable snapshots.

    Supported by --os-volume-api-version 3.8 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        host_group = parser.add_mutually_exclusive_group()
        host_group.add_argument(
            "host",
            metavar="<host>",
            nargs='?',
            help=_(
                'Cinder host on which to list manageable snapshots. '
                'Takes the form: host@backend-name#pool'
            ),
        )
        host_group.add_argument(
            "--cluster",
            metavar="<cluster>",
            help=_(
                'Cinder cluster on which to list manageable snapshots. '
                'Takes the form: cluster@backend-name#pool. '
                '(supported by --os-volume-api-version 3.17 or later)'
            ),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        # TODO(stephenfin): Remove this in a future major version bump
        parser.add_argument(
            '--detailed',
            metavar='<detailed>',
            default=None,
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            '--marker',
            metavar='<marker>',
            default=None,
            help=_(
                'Begin returning snapshots that appear later in the '
                'snapshot list than that represented by this reference. '
                'This reference should be json like. Default=None.'
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            default=None,
            help=_('Maximum number of snapshots to return. Default=None.'),
        )
        parser.add_argument(
            '--offset',
            metavar='<offset>',
            default=None,
            help=_('Number of snapshots to skip after marker. Default=None.'),
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            default=None,
            help=(
                _(
                    'Comma-separated list of sort keys and directions in the '
                    'form of <key>[:<asc|desc>]. '
                    'Valid keys: %s. '
                    'Default=None.'
                )
                % ', '.join(SORT_MANAGEABLE_KEY_VALUES)
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[list[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if parsed_args.host is None and parsed_args.cluster is None:
            msg = _(
                "Either <host> or '--cluster <cluster>' needs to be provided "
                "to run the 'block storage volume snapshot manageable list' "
                "command"
            )
            raise exceptions.CommandError(msg)

        if not sdk_utils.supports_microversion(volume_client, '3.8'):
            msg = _(
                "--os-volume-api-version 3.8 or greater is required to "
                "support the 'block storage volume snapshot manageable list' "
                "command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.cluster:
            if not sdk_utils.supports_microversion(volume_client, '3.17'):
                msg = _(
                    "--os-volume-api-version 3.17 or greater is required to "
                    "support the '--cluster' option"
                )
                raise exceptions.CommandError(msg)

        detailed = parsed_args.long
        if parsed_args.detailed is not None:
            detailed = parsed_args.detailed.lower().strip() in {
                '1',
                't',
                'true',
                'on',
                'y',
                'yes',
            }
            if detailed:
                # if the user requested e.g. '--detailed true' then they should
                # not request '--long'
                msg = _(
                    "The --detailed option has been deprecated. "
                    "Use --long instead."
                )
                self.log.warning(msg)
            else:
                # if the user requested e.g. '--detailed false' then they
                # should simply stop requesting this since the default has
                # changed
                msg = _("The --detailed option has been deprecated. Unset it.")
                self.log.warning(msg)

        columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
        ]
        if detailed:
            columns.extend(
                [
                    'reason_not_safe',
                    'cinder_id',
                    'extra_info',
                ]
            )

        query: dict[str, Any] = {}
        if parsed_args.host:
            query['host'] = parsed_args.host
        if parsed_args.cluster:
            query['cluster'] = parsed_args.cluster
        if parsed_args.marker:
            query['marker'] = parsed_args.marker
        if parsed_args.limit:
            query['limit'] = parsed_args.limit
        if parsed_args.offset:
            query['offset'] = parsed_args.offset
        if parsed_args.sort:
            query['sort'] = parsed_args.sort

        data = volume_client.manageable_snapshots(details=detailed, **query)

        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )



#    Copyright 2017 FUJITSU LIMITED
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import identity as identity_utils
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network.v2.vpnaas import utils as vpn_utils


_attr_map = [
    ('id', 'ID', column_util.LIST_BOTH),
    ('name', 'Name', column_util.LIST_BOTH),
    ('auth_algorithm', 'Authentication Algorithm', column_util.LIST_BOTH),
    ('encapsulation_mode', 'Encapsulation Mode', column_util.LIST_BOTH),
    ('transform_protocol', 'Transform Protocol', column_util.LIST_BOTH),
    ('encryption_algorithm', 'Encryption Algorithm', column_util.LIST_BOTH),
    ('pfs', 'Perfect Forward Secrecy (PFS)', column_util.LIST_LONG_ONLY),
    ('description', 'Description', column_util.LIST_LONG_ONLY),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
    ('lifetime', 'Lifetime', column_util.LIST_LONG_ONLY),
]

_attr_map_dict = {
    'id': 'ID',
    'name': 'Name',
    'auth_algorithm': 'Authentication Algorithm',
    'encapsulation_mode': 'Encapsulation Mode',
    'transform_protocol': 'Transform Protocol',
    'encryption_algorithm': 'Encryption Algorithm',
    'pfs': 'Perfect Forward Secrecy (PFS)',
    'lifetime': 'Lifetime',
    'description': 'Description',
    'project_id': 'Project',
}

_auth_algorithms = [
    'sha1',
    'sha256',
    'sha384',
    'sha512',
    'aes-xcbc',
    'aes-cmac',
]

_encryption_algorithms = [
    '3des',
    'aes-128',
    'aes-192',
    'aes-256',
    'aes-128-ccm-8',
    'aes-192-ccm-8',
    'aes-256-ccm-8',
    'aes-128-ccm-12',
    'aes-192-ccm-12',
    'aes-256-ccm-12',
    'aes-128-ccm-16',
    'aes-192-ccm-16',
    'aes-256-ccm-16',
    'aes-128-gcm-8',
    'aes-192-gcm-8',
    'aes-256-gcm-8',
    'aes-128-gcm-12',
    'aes-192-gcm-12',
    'aes-256-gcm-12',
    'aes-128-gcm-16',
    'aes-192-gcm-16',
    'aes-256-gcm-16',
    'aes-128-ctr',
    'aes-192-ctr',
    'aes-256-ctr',
]

_pfs_groups = [
    'group2',
    'group5',
    'group14',
    'group15',
    'group16',
    'group17',
    'group18',
    'group19',
    'group20',
    'group21',
    'group22',
    'group23',
    'group24',
    'group25',
    'group26',
    'group27',
    'group28',
    'group29',
    'group30',
    'group31',
]


def _convert_to_lowercase(string: str) -> str:
    return string.lower()


def _get_common_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '--description',
        metavar='<description>',
        help=_('Description of the IPsec policy'),
    )
    parser.add_argument(
        '--auth-algorithm',
        choices=_auth_algorithms,
        type=_convert_to_lowercase,
        help=_('Authentication algorithm for IPsec policy'),
    )
    parser.add_argument(
        '--encapsulation-mode',
        choices=['tunnel', 'transport'],
        type=_convert_to_lowercase,
        help=_('Encapsulation mode for IPsec policy'),
    )
    parser.add_argument(
        '--encryption-algorithm',
        choices=_encryption_algorithms,
        type=_convert_to_lowercase,
        help=_('Encryption algorithm for IPsec policy'),
    )
    parser.add_argument(
        '--lifetime',
        metavar='units=<units>,value=<value>',
        action=parseractions.MultiKeyValueAction,
        optional_keys=['units', 'value'],
        help=_(
            "IPsec lifetime attributes. "
            "'units'-seconds, default:seconds. "
            "'value'-non negative integer, default:3600."
        ),
    )
    parser.add_argument(
        '--pfs',
        choices=_pfs_groups,
        type=_convert_to_lowercase,
        help=_('Perfect Forward Secrecy for IPsec policy'),
    )
    parser.add_argument(
        '--transform-protocol',
        type=_convert_to_lowercase,
        choices=['esp', 'ah', 'ah-esp'],
        help=_('Transform protocol for IPsec policy'),
    )


def _get_common_attrs(parsed_args: argparse.Namespace) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    if parsed_args.description:
        attrs['description'] = str(parsed_args.description)
    if parsed_args.auth_algorithm:
        attrs['auth_algorithm'] = parsed_args.auth_algorithm
    if parsed_args.encapsulation_mode:
        attrs['encapsulation_mode'] = parsed_args.encapsulation_mode
    if parsed_args.transform_protocol:
        attrs['transform_protocol'] = parsed_args.transform_protocol
    if parsed_args.encryption_algorithm:
        attrs['encryption_algorithm'] = parsed_args.encryption_algorithm
    if parsed_args.pfs:
        attrs['pfs'] = parsed_args.pfs
    if parsed_args.lifetime:
        vpn_utils.validate_lifetime_dict(parsed_args.lifetime)
        attrs['lifetime'] = parsed_args.lifetime
    return attrs


class CreateIPsecPolicy(command.ShowOne):
    _description = _("Create an IPsec policy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        parser.add_argument(
            'name', metavar='<name>', help=_('Name of the IPsec policy')
        )
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_common_attrs(parsed_args)
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id
        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        obj = client.create_vpn_ipsec_policy(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj,
            _attr_map_dict,
            [
                'location',
                'tenant_id',
                'phase1_negotiation_mode',
                'units',
                'value',
            ],
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteIPsecPolicy(command.Command):
    _description = _("Delete IPsec policy(policies)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ipsecpolicy',
            metavar='<ipsec-policy>',
            nargs='+',
            help=_('ipsec policy to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0
        for ipsec in parsed_args.ipsecpolicy:
            try:
                ipsec_id = client.find_vpn_ipsec_policy(
                    ipsec, ignore_missing=False
                )['id']
                client.delete_vpn_ipsec_policy(ipsec_id)
            except Exception as e:
                result += 1
                print(
                    f"Failed to delete IPsec policy with "
                    f"name or ID {ipsec}: {e}"
                )

        if result > 0:
            total = len(parsed_args.ipsecpolicy)
            msg = _(
                "%(result)s of %(total)s IPsec policy failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListIPsecPolicy(command.Lister):
    _description = _("List IPsec policies that belong to a given project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        obj = client.vpn_ipsec_policies()
        headers, columns = column_util.get_column_definitions(
            _attr_map, long_listing=parsed_args.long
        )
        return (headers, (utils.get_dict_properties(s, columns) for s in obj))


class SetIPsecPolicy(command.Command):
    _description = _("Set IPsec policy properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        parser.add_argument(
            '--name', metavar='<name>', help=_('Name of the IPsec policy')
        )
        parser.add_argument(
            'ipsecpolicy',
            metavar='<ipsec-policy>',
            help=_('IPsec policy to set (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        attrs = _get_common_attrs(parsed_args)
        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        ipsec_id = client.find_vpn_ipsec_policy(
            parsed_args.ipsecpolicy, ignore_missing=False
        )['id']
        try:
            client.update_vpn_ipsec_policy(ipsec_id, **attrs)
        except Exception as e:
            msg = _("Failed to set IPsec policy '%(ipsec)s': %(e)s") % {
                'ipsec': parsed_args.ipsecpolicy,
                'e': e,
            }
            raise exceptions.CommandError(msg)


class ShowIPsecPolicy(command.ShowOne):
    _description = _("Display IPsec policy details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ipsecpolicy',
            metavar='<ipsec-policy>',
            help=_('IPsec policy to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_vpn_ipsec_policy(
            parsed_args.ipsecpolicy, ignore_missing=False
        )
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj,
            _attr_map_dict,
            [
                'location',
                'tenant_id',
                'phase1_negotiation_mode',
                'units',
                'value',
            ],
        )
        data = utils.get_dict_properties(obj, columns)
        return (display_columns, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Availability Zone action implementations"""

import argparse
import copy
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _xform_compute_availability_zone(
    az: Any, include_extra: bool
) -> list[dict[str, str]]:
    result: list[dict[str, str]] = []
    zone_info = {
        'zone_name': az.name,
        'zone_status': (
            'available' if az.state['available'] else 'not available'
        ),
    }

    if not include_extra:
        result.append(zone_info)
        return result

    if az.hosts:
        for host, services in az.hosts.items():
            host_info = copy.deepcopy(zone_info)
            host_info['host_name'] = host

            for svc, state in services.items():
                info = copy.deepcopy(host_info)
                info['service_name'] = svc
                info['service_status'] = '{} {} {}'.format(
                    'enabled' if state['active'] else 'disabled',
                    ':-)' if state['available'] else 'XXX',
                    state['updated_at'],
                )
                result.append(info)
    else:
        zone_info['host_name'] = ''
        zone_info['service_name'] = ''
        zone_info['service_status'] = ''
        result.append(zone_info)
    return result


def _xform_share_availability_zone(az: Any) -> list[dict[str, str]]:
    result = []
    zone_info = {
        'zone_name': az.name,
        # manila doesn't have the concept of availability zone statuses so we
        # show zones as always "available"
        'zone_status': 'available',
    }
    result.append(zone_info)
    return result


def _xform_network_availability_zone(az: Any) -> list[dict[str, str]]:
    result: list[dict[str, str]] = []
    zone_info = {}
    zone_info['zone_name'] = az.name
    zone_info['zone_status'] = az.state
    if 'unavailable' == zone_info['zone_status']:
        zone_info['zone_status'] = 'not available'
    zone_info['zone_resource'] = az.resource
    result.append(zone_info)
    return result


def _xform_volume_availability_zone(az: Any) -> list[dict[str, str]]:
    result: list[dict[str, str]] = []
    zone_info = {
        'zone_name': az.name,
        'zone_status': (
            'available' if az.state['available'] else 'not available'
        ),
    }
    result.append(zone_info)
    return result


class ListAvailabilityZone(command.Lister):
    _description = _("List availability zones and their status")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--compute',
            action='store_true',
            default=False,
            help=_('List compute availability zones'),
        )
        parser.add_argument(
            '--network',
            action='store_true',
            default=False,
            help=_('List network availability zones'),
        )
        parser.add_argument(
            '--volume',
            action='store_true',
            default=False,
            help=_('List volume availability zones'),
        )
        parser.add_argument(
            '--share',
            action='store_true',
            default=False,
            help=_('List share availability zones'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def _get_compute_availability_zones(
        self, parsed_args: argparse.Namespace
    ) -> list[dict[str, str]]:
        compute_client = self.app.client_manager.compute
        try:
            data = list(compute_client.availability_zones(details=True))
        except sdk_exceptions.ForbiddenException:  # policy doesn't allow
            try:
                data = list(compute_client.availability_zones(details=False))
            except Exception:
                raise

        result = []
        for zone in data:
            result += _xform_compute_availability_zone(zone, parsed_args.long)
        return result

    def _get_network_availability_zones(
        self, parsed_args: argparse.Namespace
    ) -> list[dict[str, str]]:
        network_client = self.app.client_manager.network
        try:
            # Verify that the extension exists.
            network_client.find_extension(
                'Availability Zone', ignore_missing=False
            )
        except Exception as e:
            LOG.debug('Network availability zone exception: ', e)
            if parsed_args.network:
                message = _(
                    "Availability zones list not supported by Network API"
                )
                LOG.warning(message)
            return []

        result = []
        for zone in network_client.availability_zones():
            result += _xform_network_availability_zone(zone)
        return result

    def _get_share_availability_zone(
        self, parsed_args: argparse.Namespace
    ) -> list[dict[str, str]]:
        data = []
        try:
            share_client = self.app.client_manager.sdk_connection.share
            data = list(share_client.availability_zones())
        except Exception as e:
            if parsed_args.share:
                message = _(
                    "Availability zones list not supported by "
                    "Shared File System API"
                )
                LOG.warning(message)
            else:
                LOG.debug('Share availability zone not available: %s', e)

        result = []
        for zone in data:
            result += _xform_share_availability_zone(zone)
        return result

    def _get_volume_availability_zones(
        self, parsed_args: argparse.Namespace
    ) -> list[dict[str, str]]:
        if not self.app.client_manager.is_volume_endpoint_enabled():
            if parsed_args.volume:
                message = _(
                    "Block Storage API is not available in the current cloud"
                )
                LOG.warning(message)
            return []

        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        data = []
        try:
            data = list(volume_client.availability_zones())
        except Exception as e:
            LOG.debug('Volume availability zone exception: %s', e)
            if parsed_args.volume:
                message = _(
                    "Availability zones list not supported by "
                    "Block Storage API"
                )
                LOG.warning(message)

        result = []
        for zone in data:
            result += _xform_volume_availability_zone(zone)
        return result

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        columns: tuple[str, ...] = ('Zone Name', 'Zone Status')
        if parsed_args.long:
            columns += (
                'Zone Resource',
                'Host Name',
                'Service Name',
                'Service Status',
            )

        # Show everything by default.
        show_all = (
            not parsed_args.compute
            and not parsed_args.network
            and not parsed_args.share
            and not parsed_args.volume
        )

        result = []
        if parsed_args.compute or show_all:
            result += self._get_compute_availability_zones(parsed_args)
        if parsed_args.network or show_all:
            result += self._get_network_availability_zones(parsed_args)
        if parsed_args.share or show_all:
            result += self._get_share_availability_zone(parsed_args)
        if parsed_args.volume or show_all:
            result += self._get_volume_availability_zones(parsed_args)

        return (
            columns,
            (utils.get_dict_properties(s, columns) for s in result),
        )



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Volume v3 consistency group snapshot action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.block_storage.v3 import (
    consistency_group_snapshot as _cg_snapshot,
)
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_consistency_group_snapshot(
    consistency_group_snapshot: _cg_snapshot.ConsistencyGroupSnapshot,
) -> tuple[Sequence[str], Iterable[Any]]:
    columns = (
        'consistencygroup_id',
        'created_at',
        'description',
        'id',
        'name',
        'status',
    )
    return columns, utils.get_item_properties(
        consistency_group_snapshot, columns
    )


class CreateConsistencyGroupSnapshot(command.ShowOne):
    _description = _("Create new consistency group snapshot.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot_name",
            metavar="<snapshot-name>",
            nargs="?",
            help=_("Name of new consistency group snapshot (default to None)"),
        )
        parser.add_argument(
            "--consistency-group",
            metavar="<consistency-group>",
            help=_(
                "Consistency group to snapshot (name or ID) "
                "(default to be the same as <snapshot-name>)"
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Description of this consistency group snapshot"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        consistency_group = parsed_args.consistency_group
        if not parsed_args.consistency_group:
            # If "--consistency-group" not specified, then consistency_group
            # will be the same as the new consistency group snapshot name
            consistency_group = parsed_args.snapshot_name
        consistency_group_id = volume_client.find_consistency_group(
            consistency_group, ignore_missing=False
        ).id
        consistency_group_snapshot = (
            volume_client.create_consistency_group_snapshot(
                consistencygroup_id=consistency_group_id,
                name=parsed_args.snapshot_name,
                description=parsed_args.description,
            )
        )

        return _format_consistency_group_snapshot(consistency_group_snapshot)


class DeleteConsistencyGroupSnapshot(command.Command):
    _description = _("Delete consistency group snapshot(s).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "consistency_group_snapshot",
            metavar="<consistency-group-snapshot>",
            nargs="+",
            help=_("Consistency group snapshot(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        for snapshot in parsed_args.consistency_group_snapshot:
            try:
                consistency_group_snapshot = (
                    volume_client.find_consistency_group_snapshot(
                        snapshot, ignore_missing=False
                    )
                )
                volume_client.delete_consistency_group_snapshot(
                    consistency_group_snapshot
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete consistency group snapshot "
                        "with name or ID '%(snapshot)s': %(e)s"
                    ),
                    {'snapshot': snapshot, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.consistency_group_snapshot)
            msg = _(
                "%(result)s of %(total)s consistency group snapshots "
                "failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListConsistencyGroupSnapshot(command.Lister):
    _description = _("List consistency group snapshots.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action="store_true",
            help=_(
                'Show detail for all projects (admin only) (defaults to False)'
            ),
        )
        parser.add_argument(
            '--long',
            action="store_true",
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--status',
            metavar="<status>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'error_deleting',
            ],
            help=_(
                'Filters results by a status ("available", "error", '
                '"creating", "deleting" or "error_deleting")'
            ),
        )
        parser.add_argument(
            '--consistency-group',
            metavar="<consistency-group>",
            help=_('Filters results by a consistency group (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if parsed_args.long:
            column_headers = [
                'ID',
                'Status',
                'ConsistencyGroup ID',
                'Name',
                'Description',
                'Created At',
            ]
            columns = [
                'id',
                'status',
                'consistencygroup_id',
                'name',
                'description',
                'created_at',
            ]
        else:
            column_headers = ['ID', 'Status', 'Name']
            columns = ['id', 'status', 'name']

        consistency_group_id = None
        if parsed_args.consistency_group:
            consistency_group_id = volume_client.find_consistency_group(
                parsed_args.consistency_group, ignore_missing=False
            ).id

        consistency_group_snapshots = (
            volume_client.consistency_group_snapshots(
                all_tenants=parsed_args.all_projects,
                status=parsed_args.status,
                consistencygroup_id=consistency_group_id,
            )
        )

        return (
            column_headers,
            (
                utils.get_item_properties(s, columns)
                for s in consistency_group_snapshots
            ),
        )


class ShowConsistencyGroupSnapshot(command.ShowOne):
    _description = _("Display consistency group snapshot details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "consistency_group_snapshot",
            metavar="<consistency-group-snapshot>",
            help=_("Consistency group snapshot to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        consistency_group_snapshot = (
            volume_client.find_consistency_group_snapshot(
                parsed_args.consistency_group_snapshot, ignore_missing=False
            )
        )
        return _format_consistency_group_snapshot(consistency_group_snapshot)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call

from openstack.block_storage.v3 import type as _type
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import volume_type


class TestTypeCreate(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.new_volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.columns = (
            'description',
            'id',
            'is_public',
            'name',
            'properties',
        )
        self.data = (
            self.new_volume_type.description,
            self.new_volume_type.id,
            self.new_volume_type.is_public,
            self.new_volume_type.name,
            format_columns.DictColumn(self.new_volume_type.extra_specs),
        )

        self.volume_client.create_type.return_value = self.new_volume_type
        self.identity_sdk_client.find_project.return_value = self.project

        self.cmd = volume_type.CreateVolumeType(self.app, None)

    def test_type_create_public(self):
        arglist = [
            "--description",
            self.new_volume_type.description,
            "--public",
            self.new_volume_type.name,
        ]
        verifylist = [
            ("description", self.new_volume_type.description),
            ("is_public", True),
            ("name", self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.create_type.assert_called_with(
            name=self.new_volume_type.name,
            description=self.new_volume_type.description,
            is_public=True,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_create_private(self):
        arglist = [
            "--description",
            self.new_volume_type.description,
            "--private",
            "--project",
            self.project.id,
            self.new_volume_type.name,
        ]
        verifylist = [
            ("description", self.new_volume_type.description),
            ("is_public", False),
            ("project", self.project.id),
            ("name", self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.create_type.assert_called_with(
            name=self.new_volume_type.name,
            description=self.new_volume_type.description,
            is_public=False,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_create_with_properties(self):
        arglist = [
            '--property',
            'myprop=myvalue',
            # this combination isn't viable server-side but is okay for testing
            '--multiattach',
            '--cacheable',
            '--replicated',
            '--availability-zone',
            'az1',
            self.new_volume_type.name,
        ]
        verifylist = [
            ('properties', {'myprop': 'myvalue'}),
            ('multiattach', True),
            ('cacheable', True),
            ('replicated', True),
            ('availability_zones', ['az1']),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result_type = sdk_fakes.generate_fake_resource(
            _type.Type,
            extra_specs={
                'myprop': 'myvalue',
                'multiattach': '<is> True',
                'cacheable': '<is> True',
                'replication_enabled': '<is> True',
                'RESKEY:availability_zones': 'az1',
            },
        )
        self.volume_client.update_type_extra_specs.return_value = result_type

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.create_type.assert_called_with(
            name=self.new_volume_type.name, description=None
        )
        self.volume_client.update_type_extra_specs.assert_called_once_with(
            self.new_volume_type.id,
            myprop='myvalue',
            multiattach='<is> True',
            cacheable='<is> True',
            replication_enabled='<is> True',
            **{'RESKEY:availability_zones': 'az1'},
        )

        expected_data = (
            self.new_volume_type.description,
            self.new_volume_type.id,
            self.new_volume_type.is_public,
            self.new_volume_type.name,
            format_columns.DictColumn(result_type.extra_specs),
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(expected_data, data)

    def test_public_type_create_with_project_public(self):
        arglist = [
            '--project',
            self.project.id,
            self.new_volume_type.name,
        ]
        verifylist = [
            ('is_public', None),
            ('project', self.project.id),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_type_create_with_encryption(self):
        encryption_type = sdk_fakes.generate_fake_resource(
            _type.TypeEncryption,
            provider='LuksEncryptor',
            cipher='aes-xts-plain64',
            key_size='128',
            control_location='front-end',
        )
        self.new_volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.volume_client.create_type.return_value = self.new_volume_type
        self.volume_client.create_type_encryption.return_value = (
            encryption_type
        )
        expected_encryption_info = {
            'provider': encryption_type.provider,
            'cipher': encryption_type.cipher,
            'key_size': encryption_type.key_size,
            'control_location': encryption_type.control_location,
            'encryption_id': encryption_type.encryption_id,
        }
        encryption_columns = (
            'description',
            'encryption',
            'id',
            'is_public',
            'name',
            'properties',
        )
        encryption_data = (
            self.new_volume_type.description,
            format_columns.DictColumn(expected_encryption_info),
            self.new_volume_type.id,
            self.new_volume_type.is_public,
            self.new_volume_type.name,
            format_columns.DictColumn(self.new_volume_type.extra_specs),
        )
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.new_volume_type.name,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.create_type.assert_called_with(
            name=self.new_volume_type.name,
            description=None,
        )
        self.volume_client.create_type_encryption.assert_called_with(
            self.new_volume_type,
            provider='LuksEncryptor',
            cipher='aes-xts-plain64',
            key_size=128,
            control_location='front-end',
        )
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, data)


class TestTypeDelete(volume_fakes.TestVolume):
    volume_types = list(sdk_fakes.generate_fake_resources(_type.Type, count=2))

    def setUp(self):
        super().setUp()

        self.volume_client.find_type.side_effect = self.volume_types
        self.volume_client.delete_type.return_value = None

        self.cmd = volume_type.DeleteVolumeType(self.app, None)

    def test_type_delete(self):
        arglist = [self.volume_types[0].id]
        verifylist = [("volume_types", [self.volume_types[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_type.assert_called_with(
            self.volume_types[0].id, ignore_missing=False
        )
        self.volume_client.delete_type.assert_called_with(self.volume_types[0])
        self.assertIsNone(result)

    def test_delete_multiple_types(self):
        arglist = []
        for t in self.volume_types:
            arglist.append(t.id)
        verifylist = [
            ('volume_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for t in self.volume_types:
            calls.append(call(t))
        self.volume_client.delete_type.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_types_with_exception(self):
        arglist = [
            self.volume_types[0].id,
            'unexist_type',
        ]
        verifylist = [
            ('volume_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.volume_client.find_type.side_effect = [
            self.volume_types[0],
            exceptions.CommandError,
        ]
        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 volume types failed to delete.', str(e))
        self.volume_client.find_type.assert_any_call(
            self.volume_types[0].id, ignore_missing=False
        )
        self.volume_client.find_type.assert_any_call(
            'unexist_type', ignore_missing=False
        )

        self.assertEqual(2, self.volume_client.find_type.call_count)
        self.volume_client.delete_type.assert_called_once_with(
            self.volume_types[0]
        )


class TestTypeList(volume_fakes.TestVolume):
    volume_types = list(sdk_fakes.generate_fake_resources(_type.Type, count=2))

    columns = [
        "ID",
        "Name",
        "Is Public",
    ]
    columns_long = [*columns, "Description"]
    data_with_default_type = [(volume_types[0].id, volume_types[0].name, True)]
    data = []
    for t in volume_types:
        data.append(
            (
                t.id,
                t.name,
                t.is_public,
            )
        )
    data_long = []
    for t in volume_types:
        data_long.append(
            (
                t.id,
                t.name,
                t.is_public,
                t.description,
            )
        )

    def setUp(self):
        super().setUp()

        self.volume_client.types.return_value = self.volume_types

        self.cmd = volume_type.ListVolumeType(self.app, None)

    def test_type_list_without_options(self):
        arglist = []
        verifylist = [
            ("long", False),
            ("is_public", 'none'),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.types.assert_called_once_with(is_public='none')
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_type_list_with_options(self):
        arglist = [
            "--long",
            "--public",
        ]
        verifylist = [
            ("long", True),
            ("is_public", True),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.types.assert_called_once_with(is_public=True)
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))

    def test_type_list_with_private_option(self):
        arglist = [
            "--private",
        ]
        verifylist = [
            ("long", False),
            ("is_public", False),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.types.assert_called_once_with(is_public=False)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_type_list_with_default_option(self):
        self.volume_client.get_type.return_value = self.volume_types[0]

        arglist = [
            "--default",
        ]
        verifylist = [
            ("encryption_type", False),
            ("long", False),
            ("is_public", 'none'),
            ("default", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.get_type.assert_called_once_with('default')
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data_with_default_type, list(data))

    def test_type_list_with_encryption(self):
        encryption_types = [
            sdk_fakes.generate_fake_resource(
                _type.TypeEncryption,
                volume_type_id=vt.id,
            )
            for vt in self.volume_types
        ]
        expected_encryption_info = [
            {
                'provider': encryption_type.provider,
                'cipher': encryption_type.cipher,
                'key_size': encryption_type.key_size,
                'control_location': encryption_type.control_location,
                'encryption_id': encryption_type.encryption_id,
            }
            for encryption_type in encryption_types
        ]
        encryption_columns = [*self.columns, "Encryption"]
        encryption_data = [
            (
                self.volume_types[x].id,
                self.volume_types[x].name,
                self.volume_types[x].is_public,
                volume_type.EncryptionInfoColumn(
                    self.volume_types[x].id,
                    {self.volume_types[x].id: expected_encryption_info[x]},
                ),
            )
            for x in (0, 1)
        ]

        self.volume_client.get_type_encryption.side_effect = encryption_types
        arglist = [
            "--encryption-type",
        ]
        verifylist = [
            ("encryption_type", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.get_type_encryption.assert_has_calls(
            [call(self.volume_types[0].id), call(self.volume_types[1].id)],
        )
        self.volume_client.types.assert_called_once_with(is_public='none')
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, list(data))


class TestTypeSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.volume_client.find_type.return_value = self.volume_type
        self.volume_client.create_type_encryption.return_value = None
        self.volume_client.update_type_encryption.return_value = None

        self.cmd = volume_type.SetVolumeType(self.app, None)

    def test_type_set(self):
        arglist = [
            '--name',
            'new_name',
            '--description',
            'new_description',
            '--private',
            self.volume_type.id,
        ]
        verifylist = [
            ('name', 'new_name'),
            ('description', 'new_description'),
            ('properties', None),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': 'new_name',
            'description': 'new_description',
            'is_public': False,
        }
        self.volume_client.update_type.assert_called_with(
            self.volume_type.id, **kwargs
        )
        self.assertIsNone(result)

        self.volume_client.add_type_access.assert_not_called()
        self.volume_client.update_type_encryption.assert_not_called()
        self.volume_client.create_type_encryption.assert_not_called()

    def test_type_set_property(self):
        arglist = [
            '--property',
            'myprop=myvalue',
            # this combination isn't viable server-side but is okay for testing
            '--multiattach',
            '--cacheable',
            '--replicated',
            '--availability-zone',
            'az1',
            self.volume_type.id,
        ]
        verifylist = [
            ('name', None),
            ('description', None),
            ('properties', {'myprop': 'myvalue'}),
            ('multiattach', True),
            ('cacheable', True),
            ('replicated', True),
            ('availability_zones', ['az1']),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.update_type_extra_specs.assert_called_once_with(
            self.volume_type.id,
            myprop='myvalue',
            multiattach='<is> True',
            cacheable='<is> True',
            replication_enabled='<is> True',
            **{'RESKEY:availability_zones': 'az1'},
        )
        self.volume_client.add_type_access.assert_not_called()
        self.volume_client.update_type_encryption.assert_not_called()
        self.volume_client.create_type_encryption.assert_not_called()

    def test_type_set_with_empty_project(self):
        arglist = [
            '--project',
            '',
            self.volume_type.id,
        ]
        verifylist = [
            ('project', ''),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.update_type_extra_specs.assert_not_called()
        self.volume_client.add_type_access.assert_not_called()
        self.volume_client.update_type_encryption.assert_not_called()
        self.volume_client.create_type_encryption.assert_not_called()

    def test_type_set_with_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.volume_type.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.update_type_extra_specs.assert_not_called()
        self.volume_client.add_type_access.assert_called_with(
            self.volume_type.id,
            self.project.id,
        )
        self.volume_client.update_type_encryption.assert_not_called()
        self.volume_client.create_type_encryption.assert_not_called()

    def test_type_set_with_new_encryption(self):
        self.volume_client.update_type_encryption.side_effect = (
            exceptions.NotFound('NotFound')
        )
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.update_type_encryption.assert_called_with(
            encryption=None,
            volume_type=self.volume_type,
            provider='LuksEncryptor',
            cipher='aes-xts-plain64',
            key_size=128,
            control_location='front-end',
        )
        self.volume_client.create_type_encryption.assert_called_with(
            self.volume_type,
            provider='LuksEncryptor',
            cipher='aes-xts-plain64',
            key_size=128,
            control_location='front-end',
        )

    def test_type_set_with_existing_encryption(self):
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.update_type_extra_specs.assert_not_called()
        self.volume_client.add_type_access.assert_not_called()
        self.volume_client.update_type_encryption.assert_called_with(
            encryption=None,
            volume_type=self.volume_type,
            provider='LuksEncryptor',
            cipher='aes-xts-plain64',
            control_location='front-end',
        )
        self.volume_client.create_type_encryption.assert_not_called()

    def test_type_set_new_encryption_without_provider(self):
        self.volume_client.update_type_encryption.side_effect = (
            exceptions.NotFound('NotFound')
        )
        arglist = [
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertEqual(
            "Command Failed: One or more of the operations failed",
            str(exc),
        )

        self.volume_client.update_type_extra_specs.assert_not_called()
        self.volume_client.add_type_access.assert_not_called()
        self.volume_client.update_type_encryption.assert_called_with(
            encryption=None,
            volume_type=self.volume_type,
            cipher='aes-xts-plain64',
            key_size=128,
            control_location='front-end',
        )
        self.volume_client.create_type_encryption.assert_not_called()


class TestTypeShow(volume_fakes.TestVolume):
    columns = (
        'access_project_ids',
        'description',
        'id',
        'is_public',
        'name',
        'properties',
    )

    def setUp(self):
        super().setUp()

        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.data = (
            None,
            self.volume_type.description,
            self.volume_type.id,
            self.volume_type.is_public,
            self.volume_type.name,
            format_columns.DictColumn(self.volume_type.extra_specs),
        )

        self.volume_client.find_type.return_value = self.volume_type

        self.cmd = volume_type.ShowVolumeType(self.app, None)

    def test_type_show(self):
        arglist = [self.volume_type.id]
        verifylist = [
            ("encryption_type", False),
            ("volume_type", self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_type.assert_called_with(
            self.volume_type.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_show_with_access(self):
        arglist = [self.volume_type.id]
        verifylist = [("volume_type", self.volume_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        private_type = sdk_fakes.generate_fake_resource(
            _type.Type, is_public=False
        )
        type_access_list = {
            'volume_type_id': private_type.id,
            'project_id': 'project-id-test',
        }
        self.volume_client.find_type.return_value = private_type
        self.volume_client.get_type_access.return_value = [type_access_list]

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_type.assert_called_once_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.get_type_access.assert_called_once_with(
            private_type.id
        )

        self.assertEqual(self.columns, columns)
        private_type_data = (
            format_columns.ListColumn([type_access_list['project_id']]),
            private_type.description,
            private_type.id,
            private_type.is_public,
            private_type.name,
            format_columns.DictColumn(private_type.extra_specs),
        )
        self.assertCountEqual(private_type_data, data)

    def test_type_show_with_list_access_exec(self):
        arglist = [self.volume_type.id]
        verifylist = [("volume_type", self.volume_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        private_type = sdk_fakes.generate_fake_resource(
            _type.Type, is_public=False
        )
        self.volume_client.find_type.return_value = private_type
        self.volume_client.get_type_access.side_effect = Exception()

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_type.assert_called_once_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.get_type_access.assert_called_once_with(
            private_type.id
        )

        self.assertEqual(self.columns, columns)
        private_type_data = (
            None,
            private_type.description,
            private_type.id,
            private_type.is_public,
            private_type.name,
            format_columns.DictColumn(private_type.extra_specs),
        )
        self.assertCountEqual(private_type_data, data)

    def test_type_show_with_encryption(self):
        encryption_type = sdk_fakes.generate_fake_resource(
            _type.TypeEncryption,
        )
        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.volume_client.find_type.return_value = self.volume_type
        self.volume_client.get_type_encryption.return_value = encryption_type
        expected_encryption_info = {
            'cipher': encryption_type.cipher,
            'control_location': encryption_type.control_location,
            'encryption_id': encryption_type.encryption_id,
            'key_size': encryption_type.key_size,
            'provider': encryption_type.provider,
        }
        encryption_columns = (
            'access_project_ids',
            'description',
            'encryption',
            'id',
            'is_public',
            'name',
            'properties',
        )
        encryption_data = (
            None,
            self.volume_type.description,
            format_columns.DictColumn(expected_encryption_info),
            self.volume_type.id,
            self.volume_type.is_public,
            self.volume_type.name,
            format_columns.DictColumn(self.volume_type.extra_specs),
        )
        arglist = ['--encryption-type', self.volume_type.id]
        verifylist = [
            ('encryption_type', True),
            ("volume_type", self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_type.assert_called_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.get_type_encryption.assert_called_with(
            self.volume_type.id
        )
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, data)


class TestTypeUnset(volume_fakes.TestVolume):
    volume_type = sdk_fakes.generate_fake_resource(_type.Type)

    def setUp(self):
        super().setUp()

        self.volume_client.find_type.return_value = self.volume_type

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.cmd = volume_type.UnsetVolumeType(self.app, None)

    def test_type_unset(self):
        arglist = [
            '--property',
            'property',
            '--property',
            'multi_property',
            self.volume_type.id,
        ]
        verifylist = [
            ('properties', ['property', 'multi_property']),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.volume_client.delete_type_extra_specs.assert_called_once_with(
            self.volume_type.id, ['property', 'multi_property']
        )
        self.assertIsNone(result)

    def test_type_unset_project_access(self):
        arglist = [
            '--project',
            self.project.id,
            self.volume_type.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.remove_type_access.assert_called_with(
            self.volume_type.id,
            self.project.id,
        )

    def test_type_unset_not_called_without_project_argument(self):
        arglist = [
            '--project',
            '',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_type', False),
            ('project', ''),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)
        self.volume_client.delete_type_encryption.assert_not_called()
        self.volume_client.remove_type_access.assert_not_called()

    def test_type_unset_failed_with_missing_volume_type_argument(self):
        arglist = [
            '--project',
            'foo',
        ]
        verifylist = [
            ('project', 'foo'),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_type_unset_encryption_type(self):
        arglist = [
            '--encryption-type',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_type', True),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.volume_client.delete_type_encryption.assert_called_with(
            None, self.volume_type.id
        )
        self.assertIsNone(result)


class TestColumns(volume_fakes.TestVolume):
    def test_encryption_info_column_with_info(self):
        fake_volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        type_id = fake_volume_type.id

        encryption_info = {
            'provider': 'LuksEncryptor',
            'cipher': None,
            'key_size': None,
            'control_location': 'front-end',
        }
        col = volume_type.EncryptionInfoColumn(
            type_id, {type_id: encryption_info}
        )
        self.assertEqual(
            utils.format_dict(encryption_info), col.human_readable()
        )
        self.assertEqual(encryption_info, col.machine_readable())

    def test_encryption_info_column_without_info(self):
        fake_volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        type_id = fake_volume_type.id

        col = volume_type.EncryptionInfoColumn(type_id, {})
        self.assertEqual('-', col.human_readable())
        self.assertIsNone(col.machine_readable())



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

# TODO(stephenfin): Remove the contents of this module in favour of the osc_lib
# version once our min version is bumped to 4.3.0

import json
from unittest import mock

from keystoneauth1 import fixture
from osc_lib.tests.fakes import (
    FakeApp,
    FakeClientManager as BaseFakeClientManager,
    FakeLog,
    FakeOptions,
    FakeResource as BaseFakeResource,
    FakeStdout,
)
import requests

__all__ = [
    'AUTH_TOKEN',
    'AUTH_URL',
    'INTERFACE',
    'PASSWORD',
    'PROJECT_NAME',
    'REGION_NAME',
    'TEST_RESPONSE_DICT',
    'TEST_RESPONSE_DICT_V3',
    'TEST_VERSIONS',
    'USERNAME',
    'VERSION',
    'FakeApp',
    'FakeClientManager',
    'FakeLog',
    'FakeOptions',
    'FakeResource',
    'FakeResponse',
    'FakeStdout',
]

AUTH_TOKEN = "foobar"
AUTH_URL = "http://0.0.0.0"
USERNAME = "itchy"
PASSWORD = "scratchy"
PROJECT_NAME = "poochie"
REGION_NAME = "richie"
INTERFACE = "catchy"
VERSION = "3"

TEST_RESPONSE_DICT = fixture.V2Token(token_id=AUTH_TOKEN, user_name=USERNAME)
_s = TEST_RESPONSE_DICT.add_service('identity', name='keystone')
_s.add_endpoint(AUTH_URL + ':5000/v2.0')
_s = TEST_RESPONSE_DICT.add_service('network', name='neutron')
_s.add_endpoint(AUTH_URL + ':9696')
_s = TEST_RESPONSE_DICT.add_service('compute', name='nova')
_s.add_endpoint(AUTH_URL + ':8774/v2.1')
_s = TEST_RESPONSE_DICT.add_service('image', name='glance')
_s.add_endpoint(AUTH_URL + ':9292')
_s = TEST_RESPONSE_DICT.add_service('object', name='swift')
_s.add_endpoint(AUTH_URL + ':8080/v1')

TEST_RESPONSE_DICT_V3 = fixture.V3Token(user_name=USERNAME)
TEST_RESPONSE_DICT_V3.set_project_scope()

TEST_VERSIONS = fixture.DiscoveryList(href=AUTH_URL)


class FakeClientManager(BaseFakeClientManager):
    _api_version = {
        'image': '2',
    }

    def __init__(self):
        super().__init__()

        self.sdk_connection = mock.Mock()

        self.network_endpoint_enabled = True
        self.compute_endpoint_enabled = True
        self.volume_endpoint_enabled = True

        # The source of configuration. This is either 'cloud_config' (a
        # clouds.yaml file) or 'global_env' ('OS_'-prefixed envvars)
        self.configuration_type = 'cloud_config'

    def get_configuration(self):
        config = {
            'region': REGION_NAME,
            'identity_api_version': VERSION,
        }

        if self.configuration_type == 'cloud_config':
            config['auth'] = {
                'username': USERNAME,
                'password': PASSWORD,
                'token': AUTH_TOKEN,
            }
        elif self.configuration_type == 'global_env':
            config['username'] = USERNAME
            config['password'] = PASSWORD
            config['token'] = AUTH_TOKEN

        return config

    def is_network_endpoint_enabled(self):
        return self.network_endpoint_enabled

    def is_compute_endpoint_enabled(self):
        return self.compute_endpoint_enabled

    def is_volume_endpoint_enabled(self, client=None):
        return self.volume_endpoint_enabled


class FakeResource(BaseFakeResource):
    def to_dict(self):
        return self._info

    @property
    def info(self):
        return self._info

    def __getitem__(self, item):
        return self._info.get(item)

    def get(self, item, default=None):
        return self._info.get(item, default)

    def pop(self, key, default_value=None):
        return self.info.pop(key, default_value)


class FakeResponse(requests.Response):
    def __init__(
        self, headers=None, status_code=200, data=None, encoding=None
    ):
        super().__init__()

        headers = headers or {}

        self.status_code = status_code

        self.headers.update(headers)
        self._content = json.dumps(data)
        if not isinstance(self._content, bytes):
            self._content = self._content.encode()



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Limits action implementations."""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.identity.v3 import limit as _limit
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as common_utils

LOG = logging.getLogger(__name__)


def _format_limit(
    limit: _limit.Limit,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    columns = (
        "description",
        "id",
        "project_id",
        "region_id",
        "resource_limit",
        "resource_name",
        "service_id",
    )
    column_headers = (
        "description",
        "id",
        "project_id",
        "region_id",
        "resource_limit",
        "resource_name",
        "service_id",
    )
    return (column_headers, utils.get_item_properties(limit, columns))


class CreateLimit(command.ShowOne):
    _description = _("Create a limit")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the limit'),
        )
        parser.add_argument(
            '--region',
            metavar='<region>',
            help=_('Region for the limit to affect.'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            required=True,
            help=_('Project to associate the resource limit to'),
        )
        common_utils.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--service',
            metavar='<service>',
            required=True,
            help=_('Service responsible for the resource to limit'),
        )
        parser.add_argument(
            '--resource-limit',
            metavar='<resource-limit>',
            required=True,
            type=int,
            help=_('The resource limit for the project to assume'),
        )
        parser.add_argument(
            'resource_name',
            metavar='<resource-name>',
            help=_('The name of the resource to limit'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {
            "resource_name": parsed_args.resource_name,
            "resource_limit": parsed_args.resource_limit,
        }
        if parsed_args.description:
            kwargs["description"] = parsed_args.description

        kwargs["project_id"] = common_utils.find_project_id_sdk(
            identity_client,
            parsed_args.project,
            domain_name_or_id=parsed_args.project_domain,
        )

        kwargs["service_id"] = common_utils.find_service_sdk(
            identity_client, parsed_args.service
        ).id

        if parsed_args.region:
            kwargs["region_id"] = identity_client.get_region(
                parsed_args.region
            ).id

        limit = identity_client.create_limit(**kwargs)

        return _format_limit(limit)


class ListLimit(command.Lister):
    _description = _("List limits")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--service',
            metavar='<service>',
            help=_('Service responsible for the resource to limit'),
        )
        parser.add_argument(
            '--resource-name',
            metavar='<resource-name>',
            dest='resource_name',
            help=_('The name of the resource to limit'),
        )
        parser.add_argument(
            '--region',
            metavar='<region>',
            help=_('Region for the registered limit to affect.'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('List resource limits associated with project'),
        )
        common_utils.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs: dict[str, object] = {}
        if parsed_args.service:
            kwargs["service_id"] = common_utils.find_service_sdk(
                identity_client, parsed_args.service
            )

        if parsed_args.region:
            kwargs["region_id"] = identity_client.get_region(
                parsed_args.region
            ).id

        if parsed_args.project:
            project_domain_id = None
            if parsed_args.project_domain:
                project_domain_id = common_utils.find_domain_id_sdk(
                    identity_client, parsed_args.project_domain
                )

            kwargs["project_id"] = common_utils._find_sdk_id(
                identity_client.find_project,
                name_or_id=parsed_args.project,
                domain_id=project_domain_id,
            )

        if parsed_args.resource_name:
            kwargs["resource_name"] = parsed_args.resource_name

        limits = identity_client.limits(**kwargs)

        columns = (
            'ID',
            'Project ID',
            'Service ID',
            'Resource Name',
            'Resource Limit',
            'Description',
            'Region ID',
        )
        return (
            columns,
            (utils.get_item_properties(s, columns) for s in limits),
        )


class ShowLimit(command.ShowOne):
    _description = _("Display limit details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'limit_id',
            metavar='<limit-id>',
            help=_('Limit to display (ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        limit = identity_client.get_limit(parsed_args.limit_id)
        return _format_limit(limit)


class SetLimit(command.ShowOne):
    _description = _("Update information about a limit")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'limit_id',
            metavar='<limit-id>',
            help=_('Limit to update (ID)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the limit'),
        )
        parser.add_argument(
            '--resource-limit',
            metavar='<resource-limit>',
            dest='resource_limit',
            type=int,
            help=_('The resource limit for the project to assume'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}
        if parsed_args.description:
            kwargs["description"] = parsed_args.description
        if parsed_args.resource_limit:
            kwargs["resource_limit"] = parsed_args.resource_limit
        limit = identity_client.update_limit(parsed_args.limit_id, **kwargs)

        return _format_limit(limit)


class DeleteLimit(command.Command):
    _description = _("Delete a limit")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'limit_id',
            metavar='<limit-id>',
            nargs="+",
            help=_(
                'Limit to delete (ID) '
                '(repeat option to remove multiple limits)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        errors = 0
        for limit_id in parsed_args.limit_id:
            try:
                identity_client.delete_limit(limit_id)
            except Exception as e:
                errors += 1
                LOG.error(
                    _("Failed to delete limit with ID '%(id)s': %(e)s"),
                    {'id': limit_id, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.limit_id)
            msg = _("%(errors)s of %(total)s limits failed to delete.") % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from osc_lib import exceptions

from openstackclient.network.v2 import l3_conntrack_helper
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestConntrackHelper(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.router = network_fakes.create_one_router()
        self.network_client.find_router.return_value = self.router


class TestCreateL3ConntrackHelper(TestConntrackHelper):
    def setUp(self):
        super().setUp()
        attrs = {'router_id': self.router.id}
        self.ct_helper = (
            network_fakes.FakeL3ConntrackHelper.create_one_l3_conntrack_helper(
                attrs
            )
        )
        self.columns = ('helper', 'id', 'port', 'protocol', 'router_id')

        self.data = (
            self.ct_helper.helper,
            self.ct_helper.id,
            self.ct_helper.port,
            self.ct_helper.protocol,
            self.ct_helper.router_id,
        )
        self.network_client.create_conntrack_helper.return_value = (
            self.ct_helper
        )

        # Get the command object to test
        self.cmd = l3_conntrack_helper.CreateConntrackHelper(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--helper',
            'tftp',
            '--protocol',
            'udp',
            '--port',
            '69',
            self.router.id,
        ]

        verifylist = [
            ('helper', 'tftp'),
            ('protocol', 'udp'),
            ('port', 69),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_conntrack_helper.assert_called_once_with(
            self.router.id, **{'helper': 'tftp', 'protocol': 'udp', 'port': 69}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_wrong_options(self):
        arglist = [
            '--protocol',
            'udp',
            '--port',
            '69',
            self.router.id,
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            None,
        )


class TestDeleteL3ConntrackHelper(TestConntrackHelper):
    def setUp(self):
        super().setUp()
        attrs = {'router_id': self.router.id}
        self.ct_helper = (
            network_fakes.FakeL3ConntrackHelper.create_one_l3_conntrack_helper(
                attrs
            )
        )
        self.network_client.delete_conntrack_helper.return_value = None

        # Get the command object to test
        self.cmd = l3_conntrack_helper.DeleteConntrackHelper(self.app, None)

    def test_delete(self):
        arglist = [self.ct_helper.router_id, self.ct_helper.id]
        verifylist = [
            ('conntrack_helper_id', [self.ct_helper.id]),
            ('router', self.ct_helper.router_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.delete_conntrack_helper.assert_called_once_with(
            self.ct_helper.id, self.router.id, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_delete_error(self):
        arglist = [self.router.id, self.ct_helper.id]
        verifylist = [
            ('conntrack_helper_id', [self.ct_helper.id]),
            ('router', self.router.id),
        ]
        self.network_client.delete_conntrack_helper.side_effect = Exception(
            'Error message'
        )
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestListL3ConntrackHelper(TestConntrackHelper):
    def setUp(self):
        super().setUp()
        attrs = {'router_id': self.router.id}
        self.conntrack_helpers = (
            network_fakes.FakeL3ConntrackHelper.create_l3_conntrack_helpers(
                attrs, count=3
            )
        )
        self.columns = (
            'ID',
            'Router ID',
            'Helper',
            'Protocol',
            'Port',
        )
        self.data = []
        for ct_helper in self.conntrack_helpers:
            self.data.append(
                (
                    ct_helper.id,
                    ct_helper.router_id,
                    ct_helper.helper,
                    ct_helper.protocol,
                    ct_helper.port,
                )
            )
        self.network_client.conntrack_helpers.return_value = (
            self.conntrack_helpers
        )

        # Get the command object to test
        self.cmd = l3_conntrack_helper.ListConntrackHelper(self.app, None)

    def test_conntrack_helpers_list(self):
        arglist = [self.router.id]
        verifylist = [
            ('router', self.router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.conntrack_helpers.assert_called_once_with(
            self.router.id
        )
        self.assertEqual(self.columns, columns)
        list_data = list(data)
        self.assertEqual(len(self.data), len(list_data))
        for index in range(len(list_data)):
            self.assertEqual(self.data[index], list_data[index])

    def test_conntrack_helpers_list_pagination(self):
        arglist = [
            '--marker',
            self.conntrack_helpers[0].id,
            '--limit',
            '1',
            self.router.id,
        ]
        verifylist = [
            ('marker', self.conntrack_helpers[0].id),
            ('limit', 1),
            ('router', self.router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.conntrack_helpers.assert_called_once_with(
            self.router.id,
            **{'marker': self.conntrack_helpers[0].id, 'limit': 1},
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestSetL3ConntrackHelper(TestConntrackHelper):
    def setUp(self):
        super().setUp()
        attrs = {'router_id': self.router.id}
        self.ct_helper = (
            network_fakes.FakeL3ConntrackHelper.create_one_l3_conntrack_helper(
                attrs
            )
        )
        self.network_client.update_conntrack_helper.return_value = None

        # Get the command object to test
        self.cmd = l3_conntrack_helper.SetConntrackHelper(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.router.id,
            self.ct_helper.id,
        ]
        verifylist = [
            ('router', self.router.id),
            ('conntrack_helper_id', self.ct_helper.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_conntrack_helper.assert_called_once_with(
            self.ct_helper.id, self.router.id
        )
        self.assertIsNone(result)

    def test_set_port(self):
        arglist = [
            self.router.id,
            self.ct_helper.id,
            '--port',
            '124',
        ]
        verifylist = [
            ('router', self.router.id),
            ('conntrack_helper_id', self.ct_helper.id),
            ('port', 124),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_conntrack_helper.assert_called_once_with(
            self.ct_helper.id, self.router.id, port=124
        )
        self.assertIsNone(result)


class TestShowL3ConntrackHelper(TestConntrackHelper):
    def setUp(self):
        super().setUp()
        attrs = {'router_id': self.router.id}
        self.ct_helper = (
            network_fakes.FakeL3ConntrackHelper.create_one_l3_conntrack_helper(
                attrs
            )
        )
        self.columns = ('helper', 'id', 'port', 'protocol', 'router_id')

        self.data = (
            self.ct_helper.helper,
            self.ct_helper.id,
            self.ct_helper.port,
            self.ct_helper.protocol,
            self.ct_helper.router_id,
        )
        self.network_client.get_conntrack_helper.return_value = self.ct_helper

        # Get the command object to test
        self.cmd = l3_conntrack_helper.ShowConntrackHelper(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_default_options(self):
        arglist = [
            self.router.id,
            self.ct_helper.id,
        ]
        verifylist = [
            ('router', self.router.id),
            ('conntrack_helper_id', self.ct_helper.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_conntrack_helper.assert_called_once_with(
            self.ct_helper.id, self.router.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy

from osc_lib.tests import utils as osc_lib_utils

from openstackclient import shell
from openstackclient.tests.unit.integ import base as test_base
from openstackclient.tests.unit import test_shell


class TestIntegV2ProjectID(test_base.TestInteg):
    def setUp(self):
        super().setUp()
        env = {
            "OS_AUTH_URL": test_base.V2_AUTH_URL,
            "OS_PROJECT_ID": test_shell.DEFAULT_PROJECT_ID,
            "OS_USERNAME": test_shell.DEFAULT_USERNAME,
            "OS_PASSWORD": test_shell.DEFAULT_PASSWORD,
            "OS_IDENTITY_API_VERSION": "2",
        }
        self.useFixture(osc_lib_utils.EnvFixture(copy.deepcopy(env)))

        self.token = test_base.make_v2_token(self.requests_mock)

    def test_project_id_env(self):
        _shell = shell.OpenStackShell()
        _shell.run("extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V2_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            test_shell.DEFAULT_PROJECT_ID,
            auth_req['auth']['tenantId'],
        )

    def test_project_id_arg(self):
        _shell = shell.OpenStackShell()
        _shell.run("--os-project-id wsx extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V2_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            "wsx",
            auth_req['auth']['tenantId'],
        )


class TestIntegV2ProjectName(test_base.TestInteg):
    def setUp(self):
        super().setUp()
        env = {
            "OS_AUTH_URL": test_base.V2_AUTH_URL,
            "OS_PROJECT_NAME": test_shell.DEFAULT_PROJECT_NAME,
            "OS_USERNAME": test_shell.DEFAULT_USERNAME,
            "OS_PASSWORD": test_shell.DEFAULT_PASSWORD,
            "OS_IDENTITY_API_VERSION": "2",
        }
        self.useFixture(osc_lib_utils.EnvFixture(copy.deepcopy(env)))

        self.token = test_base.make_v2_token(self.requests_mock)

    def test_project_name_env(self):
        _shell = shell.OpenStackShell()
        _shell.run("extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V2_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            test_shell.DEFAULT_PROJECT_NAME,
            auth_req['auth']['tenantName'],
        )

    def test_project_name_arg(self):
        _shell = shell.OpenStackShell()
        _shell.run("--os-project-name qaz extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V2_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            "qaz",
            auth_req['auth']['tenantName'],
        )


class TestIntegV3ProjectID(test_base.TestInteg):
    def setUp(self):
        super().setUp()
        env = {
            "OS_AUTH_URL": test_base.V3_AUTH_URL,
            "OS_PROJECT_ID": test_shell.DEFAULT_PROJECT_NAME,
            # "OS_PROJECT_DOMAIN_ID": test_shell.DEFAULT_PROJECT_DOMAIN_ID,
            # "OS_USER_DOMAIN_ID": test_shell.DEFAULT_USER_DOMAIN_ID,
            "OS_USERNAME": test_shell.DEFAULT_USERNAME,
            "OS_PASSWORD": test_shell.DEFAULT_PASSWORD,
            "OS_IDENTITY_API_VERSION": "3",
        }
        self.useFixture(osc_lib_utils.EnvFixture(copy.deepcopy(env)))

        self.token = test_base.make_v3_token(self.requests_mock)

    def test_project_id_env(self):
        _shell = shell.OpenStackShell()
        _shell.run("extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V3_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertIsNone(auth_req['auth'].get('tenantId', None))
        self.assertIsNone(auth_req['auth'].get('tenantName', None))

    def test_project_id_arg(self):
        _shell = shell.OpenStackShell()
        _shell.run("--os-project-id wsx extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V3_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertIsNone(auth_req['auth'].get('tenantId', None))
        self.assertIsNone(auth_req['auth'].get('tenantName', None))


class TestIntegV3ProjectName(test_base.TestInteg):
    def setUp(self):
        super().setUp()
        env = {
            "OS_AUTH_URL": test_base.V3_AUTH_URL,
            "OS_PROJECT_NAME": test_shell.DEFAULT_PROJECT_NAME,
            # "OS_PROJECT_DOMAIN_ID": test_shell.DEFAULT_PROJECT_DOMAIN_ID,
            # "OS_USER_DOMAIN_ID": test_shell.DEFAULT_USER_DOMAIN_ID,
            "OS_USERNAME": test_shell.DEFAULT_USERNAME,
            "OS_PASSWORD": test_shell.DEFAULT_PASSWORD,
            "OS_IDENTITY_API_VERSION": "3",
        }
        self.useFixture(osc_lib_utils.EnvFixture(copy.deepcopy(env)))

        self.token = test_base.make_v3_token(self.requests_mock)

    def test_project_name_env(self):
        _shell = shell.OpenStackShell()
        _shell.run("extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V3_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            test_shell.DEFAULT_PROJECT_NAME,
            auth_req['auth']['scope']['project']['name'],
        )

        self.assertIsNone(auth_req['auth'].get('tenantId', None))
        self.assertIsNone(auth_req['auth'].get('tenantName', None))

    def test_project_name_arg(self):
        _shell = shell.OpenStackShell()
        _shell.run("--os-project-name wsx extension list".split())

        # Check general calls
        self.assertNotEqual(len(self.requests_mock.request_history), 0)

        # Check discovery request
        self.assertEqual(
            test_base.V3_AUTH_URL,
            self.requests_mock.request_history[0].url,
        )

        # Check auth request
        auth_req = self.requests_mock.request_history[1].json()

        self.assertEqual(
            "wsx",
            auth_req['auth']['scope']['project']['name'],
        )

        self.assertIsNone(auth_req['auth'].get('tenantId', None))
        self.assertIsNone(auth_req['auth'].get('tenantName', None))



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.block_storage.v3 import capabilities as _capabilities
from openstack.block_storage.v3 import stats as _stats
from osc_lib.cli import format_columns

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_backend


def _create_fake_capability():
    """Create a fake volume backend capability.

    :param dict attrs:
        A dictionary with all attributes of the Capabilities.
    :return:
        A FakeResource object with capability name and attrs.
    """
    # Set default attribute
    return _capabilities.Capabilities(
        namespace="OS::Storage::Capabilities::fake",
        vendor_name="OpenStack",
        volume_backend_name="lvmdriver-1",
        pool_name="pool",
        driver_version="2.0.0",
        storage_protocol="iSCSI",
        display_name="Capabilities of Cinder LVM driver",
        description="Blah, blah.",
        visibility="public",
        replication_targets=[],
        properties={
            "compression": {
                "title": "Compression",
                "description": "Enables compression.",
                "type": "boolean",
            },
            "qos": {
                "title": "QoS",
                "description": "Enables QoS.",
                "type": "boolean",
            },
            "replication": {
                "title": "Replication",
                "description": "Enables replication.",
                "type": "boolean",
            },
            "thin_provisioning": {
                "title": "Thin Provisioning",
                "description": "Sets thin provisioning.",
                "type": "boolean",
            },
        },
    )


def _create_fake_pool():
    return _stats.Pools(
        name='host@lvmdriver-1#lvmdriver-1',
        capabilities={
            'storage_protocol': 'iSCSI',
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
            'total_volumes': 99,
            'total_capacity_gb': 1000.00,
            'allocated_capacity_gb': 100,
            'max_over_subscription_ratio': 200.0,
        },
    )


class TestShowVolumeCapability(volume_fakes.TestVolume):
    """Test backend capability functionality."""

    def setUp(self):
        super().setUp()

        self.capability = _create_fake_capability()
        self.volume_client.get_capabilities.return_value = self.capability

        self.cmd = volume_backend.ShowCapability(self.app, None)

    def test_capability_show(self):
        self.set_volume_api_version('3.0')

        arglist = [
            'fake',
        ]
        verifylist = [
            ('host', 'fake'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Title',
            'Key',
            'Type',
            'Description',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        capabilities = [
            'Compression',
            'Replication',
            'QoS',
            'Thin Provisioning',
        ]

        # confirming if all expected values are present in the result.
        for cap in data:
            self.assertIn(cap[0], capabilities)

        # checking if proper call was made to get capabilities
        self.volume_client.get_capabilities.assert_called_with(
            'fake',
        )


class TestListVolumePool(volume_fakes.TestVolume):
    """Tests for volume backend pool listing."""

    def setUp(self):
        super().setUp()

        self.pool = _create_fake_pool()
        self.volume_client.backend_pools.return_value = [self.pool]

        # Get the command object to test
        self.cmd = volume_backend.ListPool(self.app, None)

    def test_pool_list(self):
        self.set_volume_api_version('3.0')

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Name',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = ((self.pool.name,),)

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to list pools
        self.volume_client.backend_pools.assert_called_with(
            detailed=False,
        )

        # checking if long columns are present in output
        self.assertNotIn("total_volumes", columns)
        self.assertNotIn("storage_protocol", columns)

    def test_service_list_with_long_option(self):
        self.set_volume_api_version('3.0')

        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Name',
            'Capabilities',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = (
            (
                self.pool.name,
                format_columns.DictColumn(self.pool.capabilities),
            ),
        )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        self.volume_client.backend_pools.assert_called_with(
            detailed=True,
        )



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from openstack.identity.v3 import project as _project
from openstack.network.v2 import vpn_service as sdk_vpn_service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2.vpnaas import vpnservice
from openstackclient.tests.unit.network.v2 import fakes as test_fakes
from openstackclient.tests.unit.network.v2.vpnaas import fakes


class TestVPNService(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self._vpnservice = fakes.VPNService().create()

        def _mock_vpnservice(*args, **kwargs):
            self.network_client.find_vpn_service.assert_called_once_with(
                self._vpnservice['id'], ignore_missing=False
            )
            return {'id': args[0]}

        self.fake_router = mock.Mock()
        self.fake_subnet = mock.Mock()
        self.network_client.find_router.return_value = self.fake_router
        self.network_client.find_subnet.return_value = self.fake_subnet
        self.fake_router.id = 'router-id-' + uuid.uuid4().hex
        self.fake_subnet.id = 'subnet-id-' + uuid.uuid4().hex

        self.network_client.find_vpn_service.side_effect = mock.Mock(
            side_effect=_mock_vpnservice
        )

        self.headers = (
            'ID',
            'Name',
            'Router',
            'Subnet',
            'Flavor',
            'State',
            'Status',
            'Description',
            'Project',
            'Ext v4 IP',
            'Ext v6 IP',
        )
        self.ordered_headers = (
            'Description',
            'Ext v4 IP',
            'Ext v6 IP',
            'Flavor',
            'ID',
            'Name',
            'Project',
            'Router',
            'State',
            'Status',
            'Subnet',
        )
        self.ordered_data = (
            self._vpnservice['description'],
            self._vpnservice['external_v4_ip'],
            self._vpnservice['external_v6_ip'],
            self._vpnservice['flavor_id'],
            self._vpnservice['id'],
            self._vpnservice['name'],
            self._vpnservice['project_id'],
            self._vpnservice['router_id'],
            self._vpnservice['admin_state_up'],
            self._vpnservice['status'],
            self._vpnservice['subnet_id'],
        )
        self.ordered_columns = (
            'description',
            'external_v4_ip',
            'external_v6_ip',
            'flavor_id',
            'id',
            'name',
            'project_id',
            'router_id',
            'admin_state_up',
            'status',
            'subnet_id',
        )


class TestCreateVPNService(TestVPNService):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.network_client.create_vpn_service.return_value = self._vpnservice
        self.mocked = self.network_client.create_vpn_service
        self.cmd = vpnservice.CreateVPNService(self.app, None)

    def test_create_with_all_params(self):
        arglist = [
            '--description',
            'my-desc',
            '--project',
            self.project.id,
            '--subnet',
            self.fake_subnet.id,
            '--router',
            self.fake_router.id,
            'my-name',
        ]
        verifylist = [
            ('description', 'my-desc'),
            ('project', self.project.id),
            ('subnet', self.fake_subnet.id),
            ('router', self.fake_router.id),
            ('name', 'my-name'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)
        self.network_client.create_vpn_service.assert_called_once_with(
            description='my-desc',
            project_id=self.project.id,
            subnet_id=self.fake_subnet.id,
            router_id=self.fake_router.id,
            name='my-name',
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertCountEqual(self.ordered_data, data)


class TestDeleteVPNService(TestVPNService):
    def setUp(self):
        super().setUp()
        self.cmd = vpnservice.DeleteVPNService(self.app, None)

    def test_delete_with_one_resource(self):
        def _mock_vpn_s(*args, **kwargs):
            return sdk_vpn_service.VpnService(**{'id': args[0]})

        self.network_client.find_vpn_service.side_effect = _mock_vpn_s

        arglist = [self._vpnservice['id']]
        verifylist = [('vpnservice', [self._vpnservice['id']])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_vpn_service.assert_called_once_with(
            self._vpnservice['id']
        )
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_vpn_s(*args, **kwargs):
            return sdk_vpn_service.VpnService(**{'id': args[0]})

        self.network_client.find_vpn_service.side_effect = _mock_vpn_s

        arglist = ['target1', 'target2']
        verifylist = [('vpnservice', ['target1', 'target2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(2, self.network_client.delete_vpn_service.call_count)
        for idx, reference in enumerate(['target1', 'target2']):
            actual = ''.join(
                self.network_client.delete_vpn_service.call_args_list[idx][0]
            )
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        arglist = ['target']
        verifylist = [('vpnservice', ['target'])]

        self.network_client.find_vpn_ipsec_policy.side_effect = [
            'target',
            exceptions.CommandError,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestListVPNService(TestVPNService):
    def setUp(self):
        super().setUp()
        self.cmd = vpnservice.ListVPNService(self.app, None)

        self.short_header = (
            'ID',
            'Name',
            'Router',
            'Subnet',
            'Flavor',
            'State',
            'Status',
        )

        self.short_data = (
            self._vpnservice['id'],
            self._vpnservice['name'],
            self._vpnservice['router_id'],
            self._vpnservice['subnet_id'],
            self._vpnservice['flavor_id'],
            self._vpnservice['admin_state_up'],
            self._vpnservice['status'],
        )

        self.network_client.vpn_services.return_value = [self._vpnservice]

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_services.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_services.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetVPNService(TestVPNService):
    def setUp(self):
        super().setUp()
        self.network_client.update_vpn_service.return_value = self._vpnservice
        self.cmd = vpnservice.SetVPNSercice(self.app, None)

    def test_set_name(self):
        arglist = [self._vpnservice['id'], '--name', 'change']
        verifylist = [
            ('vpnservice', self._vpnservice['id']),
            ('name', 'change'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_service.assert_called_once_with(
            self._vpnservice['id'], name='change'
        )
        self.assertIsNone(result)

    def test_set_description(self):
        arglist = [self._vpnservice['id'], '--description', 'change-desc']
        verifylist = [
            ('vpnservice', self._vpnservice['id']),
            ('description', 'change-desc'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_service.assert_called_once_with(
            self._vpnservice['id'], description='change-desc'
        )
        self.assertIsNone(result)


class TestShowVPNService(TestVPNService):
    def setUp(self):
        super().setUp()
        self.network_client.find_vpn_service.side_effect = None
        self.network_client.find_vpn_service.return_value = self._vpnservice
        self.cmd = vpnservice.ShowVPNService(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        arglist = [self._vpnservice['id']]
        verifylist = [('vpnservice', self._vpnservice['id'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_vpn_service.assert_called_once_with(
            self._vpnservice['id'], ignore_missing=False
        )
        self.assertEqual(self.ordered_headers, headers)
        self.assertCountEqual(self.ordered_data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""L3 Conntrack Helper action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import l3_conntrack_helper as _l3_conntrack_helper
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _l3_conntrack_helper.ConntrackHelper,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _get_attrs(client: Any, parsed_args: argparse.Namespace) -> dict[str, Any]:
    router = client.find_router(parsed_args.router, ignore_missing=False)
    attrs: dict[str, Any] = {'router_id': router.id}
    if parsed_args.helper:
        attrs['helper'] = parsed_args.helper
    if parsed_args.protocol:
        attrs['protocol'] = parsed_args.protocol
    if parsed_args.port:
        attrs['port'] = parsed_args.port

    return attrs


class CreateConntrackHelper(command.ShowOne):
    _description = _("Create a new L3 conntrack helper")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_('Router for which conntrack helper will be created'),
        )
        parser.add_argument(
            '--helper',
            required=True,
            metavar='<helper>',
            help=_('The netfilter conntrack helper module'),
        )
        parser.add_argument(
            '--protocol',
            required=True,
            metavar='<protocol>',
            help=_(
                'The network protocol for the netfilter conntrack target rule'
            ),
        )
        parser.add_argument(
            '--port',
            required=True,
            metavar='<port>',
            type=int,
            help=_('The network port for the netfilter conntrack target rule'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network

        attrs = _get_attrs(client, parsed_args)
        obj = client.create_conntrack_helper(attrs.pop('router_id'), **attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteConntrackHelper(command.Command):
    _description = _("Delete L3 conntrack helper")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_('Router that the conntrack helper belongs to'),
        )
        parser.add_argument(
            'conntrack_helper_id',
            metavar='<conntrack-helper-id>',
            nargs='+',
            help=_('The ID of the conntrack helper(s) to delete'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        router = client.find_router(parsed_args.router, ignore_missing=False)
        for ct_helper in parsed_args.conntrack_helper_id:
            try:
                client.delete_conntrack_helper(
                    ct_helper, router.id, ignore_missing=False
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete L3 conntrack helper with "
                        "ID '%(ct_helper)s': %(e)s"
                    ),
                    {'ct_helper': ct_helper, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.conntrack_helper_id)
            msg = _(
                "%(result)s of %(total)s L3 conntrack helpers failed "
                "to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListConntrackHelper(command.Lister):
    _description = _("List L3 conntrack helpers")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_('Router that the conntrack helper belongs to'),
        )
        parser.add_argument(
            '--helper',
            metavar='<helper>',
            help=_(
                'List only helpers using the specified netfilter conntrack '
                'helper module'
            ),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            help=_(
                'List only helpers with the specified network protocol for '
                'the netfilter conntrack target rule'
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_(
                'List only helpers with the specified network port for '
                'the netfilter conntrack target rule (name or ID)'
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        columns = (
            'id',
            'router_id',
            'helper',
            'protocol',
            'port',
        )
        column_headers = (
            'ID',
            'Router ID',
            'Helper',
            'Protocol',
            'Port',
        )

        attrs = _get_attrs(client, parsed_args)
        if parsed_args.marker is not None:
            attrs['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            attrs['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            attrs['max_items'] = parsed_args.max_items

        data = client.conntrack_helpers(attrs.pop('router_id'), **attrs)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetConntrackHelper(command.Command):
    _description = _("Set L3 conntrack helper properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_('Router that the conntrack helper belongs to'),
        )
        parser.add_argument(
            'conntrack_helper_id',
            metavar='<conntrack-helper-id>',
            help=_('The ID of the conntrack helper(s)'),
        )
        parser.add_argument(
            '--helper',
            metavar='<helper>',
            help=_('The netfilter conntrack helper module'),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            help=_(
                'The network protocol for the netfilter conntrack target rule'
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            type=int,
            help=_('The network port for the netfilter conntrack target rule'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        attrs = _get_attrs(client, parsed_args)
        if attrs:
            client.update_conntrack_helper(
                parsed_args.conntrack_helper_id,
                attrs.pop('router_id'),
                **attrs,
            )


class ShowConntrackHelper(command.ShowOne):
    _description = _("Display L3 conntrack helper details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_('Router that the conntrack helper belongs to'),
        )
        parser.add_argument(
            'conntrack_helper_id',
            metavar='<conntrack-helper-id>',
            help=_('The ID of the conntrack helper'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        router = client.find_router(parsed_args.router, ignore_missing=False)
        obj = client.get_conntrack_helper(
            parsed_args.conntrack_helper_id, router.id
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Security Group Rule action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.network.v2 import security_group_rule as _security_group_rule
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common
from openstackclient.network import utils as network_utils

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _security_group_rule.SecurityGroupRule,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'name', 'tenant_id', 'tags']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateSecurityGroupRule(
    command.ShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create a new security group rule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_("Create rule in this security group (name or ID)"),
        )
        remote_group = parser.add_mutually_exclusive_group()
        remote_group.add_argument(
            "--remote-ip",
            metavar="<ip-address>",
            help=_(
                "Remote IP address block (may use CIDR notation; "
                "default for IPv4 rule: 0.0.0.0/0, "
                "default for IPv6 rule: ::/0)"
            ),
        )
        remote_group.add_argument(
            "--remote-group",
            metavar="<group>",
            help=_("Remote security group (name or ID)"),
        )
        remote_group.add_argument(
            "--remote-address-group",
            metavar="<group>",
            help=_("Remote address group (name or ID)"),
        )
        parser.add_argument(
            '--dst-port',
            metavar='<port-range>',
            action=parseractions.RangeAction,
            help=_(
                "Destination port, may be a single port or a starting and "
                "ending port range: 137:139. Required for IP protocols TCP "
                "and UDP. Ignored for ICMP IP protocols."
            ),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            type=network_utils.convert_to_lowercase,
            help=_(
                "IP protocol (ah, dccp, egp, esp, gre, icmp, igmp, "
                "ipv6-encap, ipv6-frag, ipv6-icmp, ipv6-nonxt, ipv6-opts, "
                "ipv6-route, ospf, pgm, rsvp, sctp, tcp, udp, udplite, vrrp "
                "and integer representations [0-255] or any; "
                "default: any (all protocols))"
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Set security group rule description"),
        )
        parser.add_argument(
            '--icmp-type',
            metavar='<icmp-type>',
            type=int,
            help=_("ICMP type for ICMP IP protocols"),
        )
        parser.add_argument(
            '--icmp-code',
            metavar='<icmp-code>',
            type=int,
            help=_("ICMP code for ICMP IP protocols"),
        )
        direction_group = parser.add_mutually_exclusive_group()
        direction_group.add_argument(
            '--ingress',
            action='store_true',
            help=_("Rule applies to incoming network traffic (default)"),
        )
        direction_group.add_argument(
            '--egress',
            action='store_true',
            help=_("Rule applies to outgoing network traffic"),
        )
        parser.add_argument(
            '--ethertype',
            metavar='<ethertype>',
            choices=['IPv4', 'IPv6'],
            type=network_utils.convert_ipvx_case,
            help=_(
                "Ethertype of network traffic "
                "(IPv4, IPv6; default: based on IP protocol)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        # Get the security group ID to hold the rule.
        security_group_id = client.find_security_group(
            parsed_args.group, ignore_missing=False
        ).id

        # Build the create attributes.
        attrs = {}
        attrs['protocol'] = network_utils.get_protocol(parsed_args)

        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description

        # NOTE(rtheis): A direction must be specified and ingress
        # is the default.
        if parsed_args.ingress or not parsed_args.egress:
            attrs['direction'] = 'ingress'
        if parsed_args.egress:
            attrs['direction'] = 'egress'

        # NOTE(rtheis): Use ethertype specified else default based
        # on IP protocol.
        attrs['ethertype'] = network_utils.get_ethertype(
            parsed_args, attrs['protocol']
        )

        # NOTE(rtheis): Validate the port range and ICMP type and code.
        # It would be ideal if argparse could do this.
        if parsed_args.dst_port and (
            parsed_args.icmp_type or parsed_args.icmp_code
        ):
            msg = _(
                'Argument --dst-port not allowed with arguments '
                '--icmp-type and --icmp-code'
            )
            raise exceptions.CommandError(msg)
        if parsed_args.icmp_type is None and parsed_args.icmp_code is not None:
            msg = _('Argument --icmp-type required with argument --icmp-code')
            raise exceptions.CommandError(msg)
        is_icmp_protocol = network_utils.is_icmp_protocol(attrs['protocol'])
        if not is_icmp_protocol and (
            parsed_args.icmp_type or parsed_args.icmp_code
        ):
            msg = _(
                'ICMP IP protocol required with arguments '
                '--icmp-type and --icmp-code'
            )
            raise exceptions.CommandError(msg)
        # NOTE(rtheis): For backwards compatibility, continue ignoring
        # the destination port range when an ICMP IP protocol is specified.
        if parsed_args.dst_port and not is_icmp_protocol:
            attrs['port_range_min'] = parsed_args.dst_port[0]
            attrs['port_range_max'] = parsed_args.dst_port[1]
        if parsed_args.icmp_type is not None and parsed_args.icmp_type >= 0:
            attrs['port_range_min'] = parsed_args.icmp_type
        if parsed_args.icmp_code is not None and parsed_args.icmp_code >= 0:
            attrs['port_range_max'] = parsed_args.icmp_code

        if parsed_args.remote_group is not None:
            attrs['remote_group_id'] = client.find_security_group(
                parsed_args.remote_group, ignore_missing=False
            ).id
        elif parsed_args.remote_address_group is not None:
            attrs['remote_address_group_id'] = client.find_address_group(
                parsed_args.remote_address_group, ignore_missing=False
            ).id
        elif parsed_args.remote_ip is not None:
            attrs['remote_ip_prefix'] = parsed_args.remote_ip
        elif attrs['ethertype'] == 'IPv4':
            attrs['remote_ip_prefix'] = '0.0.0.0/0'
        elif attrs['ethertype'] == 'IPv6':
            attrs['remote_ip_prefix'] = '::/0'
        attrs['security_group_id'] = security_group_id
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        # Create and show the security group rule.
        obj = client.create_security_group_rule(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)


class DeleteSecurityGroupRule(command.Command):
    _description = _("Delete security group rule(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rule',
            metavar='<rule>',
            nargs="+",
            help=_("Security group rule(s) to delete (ID only)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for rule in parsed_args.rule:
            try:
                obj = client.find_security_group_rule(
                    rule, ignore_missing=False
                )
                client.delete_security_group_rule(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete security group rule with "
                        "name or ID '%(rule)s': %(e)s"
                    ),
                    {'rule': rule, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.rule)
            msg = _("%(result)s of %(total)s rules failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListSecurityGroupRule(command.Lister):
    _description = _("List security group rules")

    @staticmethod
    def _format_network_security_group_rule(
        rule: _security_group_rule.SecurityGroupRule,
    ) -> dict[str, object]:
        """Transform the SDK SecurityGroupRule object to a dict

        The SDK object gets in the way of reformatting columns...
        Create port_range column from port_range_min and port_range_max
        """
        data = rule.to_dict()
        data['port_range'] = network_utils.format_network_port_range(data)
        data['remote_ip_prefix'] = network_utils.format_remote_ip_prefix(data)
        return data

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            nargs='?',
            help=_("List all rules in this security group (name or ID)"),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            type=network_utils.convert_to_lowercase,
            help=_(
                "List only rules with the specified IP protocol "
                "(ah, dhcp, egp, esp, gre, icmp, igmp, ipv6-encap, "
                "ipv6-frag, ipv6-icmp, ipv6-nonxt, ipv6-opts, ipv6-route, "
                "ospf, pgm, rsvp, sctp, tcp, udp, udplite, vrrp and integer "
                "representations [0-255] or any; "
                "default: any (all protocols))"
            ),
        )
        parser.add_argument(
            '--ethertype',
            metavar='<ethertype>',
            type=network_utils.convert_to_lowercase,
            help=_(
                "List only rules with the specified Ethertype (IPv4 or IPv6)"
            ),
        )
        direction_group = parser.add_mutually_exclusive_group()
        direction_group.add_argument(
            '--ingress',
            action='store_true',
            help=_("List only rules applied to incoming network traffic"),
        )
        direction_group.add_argument(
            '--egress',
            action='store_true',
            help=_("List only rules applied to outgoing network traffic"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("**Deprecated** This argument is no longer needed"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("List only rules with the specified project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def _get_column_headers(
        self, parsed_args: argparse.Namespace
    ) -> tuple[str, ...]:
        column_headers: tuple[str, ...] = (
            'ID',
            'IP Protocol',
            'Ethertype',
            'IP Range',
            'Port Range',
            'Direction',
            'Remote Security Group',
            'Remote Address Group',
        )
        if parsed_args.group is None:
            column_headers += ('Security Group',)
        return column_headers

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        if parsed_args.long:
            msg = _(
                "The --long option has been deprecated and is no longer needed"
            )
            self.log.warning(msg)

        column_headers = self._get_column_headers(parsed_args)
        columns: tuple[str, ...] = (
            'id',
            'protocol',
            'ether_type',
            'remote_ip_prefix',
            'port_range',
            'direction',
            'remote_group_id',
            'remote_address_group_id',
        )

        # Get the security group rules using the requested query.
        query = {}
        if parsed_args.group is not None:
            # NOTE(rtheis): Unfortunately, the security group resource
            # does not contain security group rules resources. So use
            # the security group ID in a query to get the resources.
            security_group_id = client.find_security_group(
                parsed_args.group, ignore_missing=False
            ).id
            query = {'security_group_id': security_group_id}
        else:
            columns += ('security_group_id',)

        if parsed_args.ingress:
            query['direction'] = 'ingress'
        if parsed_args.egress:
            query['direction'] = 'egress'
        if parsed_args.protocol is not None:
            query['protocol'] = parsed_args.protocol
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            query['project_id'] = project_id

        if parsed_args.marker is not None:
            query['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            query['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            query['max_items'] = parsed_args.max_items

        rules = [
            self._format_network_security_group_rule(r)
            for r in client.security_group_rules(**query)
        ]

        return (
            column_headers,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                )
                for s in rules
            ),
        )


class ShowSecurityGroupRule(command.ShowOne):
    _description = _("Display security group rule details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rule',
            metavar="<rule>",
            help=_("Security group rule to display (ID only)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_security_group_rule(
            parsed_args.rule, ignore_missing=False
        )
        # necessary for old rules that have None in this field
        if not obj['remote_ip_prefix']:
            obj['remote_ip_prefix'] = network_utils.format_remote_ip_prefix(
                obj
            )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.block_storage.v2 import service as _service
from openstack.test import fakes as sdk_fakes

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import volume_host


class TestVolumeHostSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()
        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.cmd = volume_host.SetVolumeHost(self.app, None)

    def test_volume_host_set_nothing(self):
        arglist = [self.service.host]
        verifylist = [('host', self.service.host)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.freeze_service.assert_not_called()
        self.volume_client.thaw_service.assert_not_called()
        self.assertIsNone(result)

    def test_volume_host_set_enable(self):
        arglist = ['--enable', self.service.host]
        verifylist = [('enable', True), ('host', self.service.host)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.thaw_service.assert_called_once_with(
            _service.Service(host=self.service.host)
        )
        self.volume_client.freeze_service.assert_not_called()
        self.assertIsNone(result)

    def test_volume_host_set_disable(self):
        arglist = ['--disable', self.service.host]
        verifylist = [('disable', True), ('host', self.service.host)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.freeze_service.assert_called_once_with(
            _service.Service(host=self.service.host)
        )
        self.volume_client.thaw_service.assert_not_called()
        self.assertIsNone(result)


class TestVolumeHostFailover(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()
        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.cmd = volume_host.FailoverVolumeHost(self.app, None)

    def test_volume_host_failover(self):
        arglist = ['--volume-backend', 'backend_test', self.service.host]
        verifylist = [
            ('volume_backend', 'backend_test'),
            ('host', self.service.host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.failover_service.assert_called_once_with(
            _service.Service(host=self.service.host),
            backend_id='backend_test',
        )
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import service_profile as _service_profile
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.network import common

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _service_profile.ServiceProfile,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        'is_enabled': 'enabled',
    }

    hidden_columns = ['location', 'name', 'tenant_id', 'project_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description
    if parsed_args.driver is not None:
        attrs['driver'] = parsed_args.driver
    if parsed_args.metainfo is not None:
        attrs['metainfo'] = parsed_args.metainfo
    if parsed_args.enable:
        attrs['enabled'] = True
    if parsed_args.disable:
        attrs['enabled'] = False

    return attrs


class CreateNetworkFlavorProfile(
    command.ShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create new network flavor profile")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar="<description>",
            help=_("Description for the flavor profile"),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_("Enable the flavor profile"),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_("Disable the flavor profile"),
        )
        parser.add_argument(
            '--driver',
            help=_(
                "Python module path to driver. This becomes "
                "required if --metainfo is missing and vice-versa."
            ),
        )
        parser.add_argument(
            '--metainfo',
            help=_(
                "Metainfo for the flavor profile. This becomes "
                "required if --driver is missing and vice-versa."
            ),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        if parsed_args.driver is None and parsed_args.metainfo is None:
            msg = _("Either --driver or --metainfo or both are required")
            raise exceptions.CommandError(msg)

        obj = client.create_service_profile(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteNetworkFlavorProfile(command.Command):
    _description = _("Delete network flavor profile")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'flavor_profile',
            metavar='<flavor-profile>',
            nargs='+',
            help=_("Flavor profile(s) to delete (ID only)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for flavor_profile in parsed_args.flavor_profile:
            try:
                obj = client.find_service_profile(
                    flavor_profile, ignore_missing=False
                )
                client.delete_service_profile(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete flavor profile with "
                        "ID '%(flavor_profile)s': %(e)s"
                    ),
                    {"flavor_profile": flavor_profile, "e": e},
                )
        if result > 0:
            total = len(parsed_args.flavor_profile)
            msg = _(
                "%(result)s of %(total)s flavor_profiles failed to delete."
            ) % {"result": result, "total": total}
            raise exceptions.CommandError(msg)


class ListNetworkFlavorProfile(command.Lister):
    _description = _("List network flavor profile(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = (
            'id',
            'driver',
            'is_enabled',
            'meta_info',
            'description',
        )
        column_headers = (
            'ID',
            'Driver',
            'Enabled',
            'Metainfo',
            'Description',
        )

        filters = {}
        if parsed_args.marker is not None:
            filters['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            filters['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            filters['max_items'] = parsed_args.max_items

        data = client.service_profiles(**filters)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class SetNetworkFlavorProfile(common.NeutronCommandWithExtraArgs):
    _description = _("Set network flavor profile properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor_profile',
            metavar="<flavor-profile>",
            help=_("Flavor profile to update (ID only)"),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            help=_("Description for the flavor profile"),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_("Enable the flavor profile"),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_("Disable the flavor profile"),
        )
        parser.add_argument(
            '--driver',
            help=_(
                "Python module path to driver. This becomes "
                "required if --metainfo is missing and vice-versa."
            ),
        )
        parser.add_argument(
            '--metainfo',
            help=_(
                "Metainfo for the flavor profile. This becomes "
                "required if --driver is missing and vice-versa."
            ),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_service_profile(
            parsed_args.flavor_profile, ignore_missing=False
        )
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        client.update_service_profile(obj, **attrs)


class ShowNetworkFlavorProfile(command.ShowOne):
    _description = _("Display network flavor profile details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor_profile',
            metavar='<flavor-profile>',
            help=_("Flavor profile to display (ID only)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_service_profile(
            parsed_args.flavor_profile, ignore_missing=False
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)
